#!/usr/bin/perl
# 
# Copyright (C) 2015 Roland Laifer 
#                    Karlsruhe Institute of Technology (KIT), 
#                    Steinbuch Centre for Computing (SCC)
#                    76128 Karlsruhe, Germany
#                    Roland.Laifer (at) kit.edu
#
# ------------------------------------------------------------------------
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------
#
# $Log: show_high_jobstats.pl,v $
# Revision 1.5  2015/11/02 13:42:02  rz54
# Made joblist uniqueue
#
# Revision 1.4  2015/10/29 15:36:41  rz54
# Added output of job ID list
#
# Revision 1.3  2015/10/27 16:42:51  rz54
# Lots of changes to better conform to good perl coding style
#
# Revision 1.2  2015/10/22 16:49:45  rz54
# added GPL, updated copyright, usage and some comments
#
# Revision 1.1  2015/10/20 14:04:12  rz54
# Initial revision
#
#
# $Id: show_high_jobstats.pl,v 1.5 2015/11/02 13:42:02 rz54 Exp $


use strict;
use warnings;

use List::MoreUtils qw(uniq);

use Getopt::Std;
our($opt_d, $opt_h, $opt_v, $opt_m, $opt_o, $opt_s, $opt_t);

# get script name
(my $myname = $0)	=~ s#^.*/##;
# associative arrays which stores sample count for job and type
undef my %job_type2count;
# associative arrays which stores sum (with options -s or -m) for job and type
undef my %job_type2sum;


sub usage {
    print "Usage: $myname [-d] [-h] [-v] [-t type] (-m number | -o number | -s number)\n";
    print "$myname collects and filters output created by Lustre jobstats from Lustre servers.\n";
    print "It is useful to check for jobs doing I/O over a high watermark.\n";
    print "For more information check Roland Laifer's talk at LAD'15, see\n";
    print "http://www.scc.kit.edu/scc/docs/Lustre/kit_lad15_20150922.pdf\n";
    print " -d            debug\n";
    print " -h            this help\n";
    print " -v            verbose\n";
    print " -m number     maximum medium size (sum divided by samples) for types write and read,\n";
    print "               in output omit values above specified number\n";
    print " -o number     minimum number of operations (samples), otherwise omited in output\n";
    print " -s number     minimum total sum for types write and read, in output omit values\n";
    print "               below specified number\n";
    print " -t type       type of operation (others omited in output)\n";
    print "               operations are e.g. open close getxattr getattr unlink rmdir rename sync\n";
    print "               write read punch (default: all types)\n";
    print "At least one of the options -m, -o, -s is required. If more than one of these options\n";
    print "is specified, all requirements must be fulfilled. These options are used to define\n";
    print "high watermarks and only jobs above these watermarks are printed.\n";
    print "Gets input from STDIN, e.g. by using pdsh.\n";
    print "Input from different nodes should not be mixed (use pdsh option -f1)\n";
    print "Example 1: pdsh -g mds -f 1 'lctl get_param *.*.job_stats' | $myname -o 10000 -t open\n";
    print "Example 2: pdsh -g oss -f 1 'lctl get_param *.*.job_stats' | $myname -s 1000000000\n";
    print "Example 3: pdsh -g oss -f 1 'lctl get_param *.*.job_stats' | $myname -m 500000\n";
    print "Example 4: pdsh -g oss -f 1 'lctl get_param *.*.job_stats' | $myname -m 500000 -s 100000000\n";
    exit 0;
}


sub parse_input {
    my $jobid;
    my @lines = <STDIN>;
    while ( my $line = shift(@lines) ) {
        chomp $line;
	# remove hostname on pdsh command output (should not harm if no pdsh is used)
	if ( $line =~ /^\S+:\s(.*)$/ ) {
	    $line = $1;
	}
	if ( $line =~ /^-\s+job_id:\s+(\S+)/ ) {
	    $jobid = $1;
	    print "Found line with jobid ($jobid): $line\n" if($opt_d);
	} elsif ( $line =~ /^\s+(\S+):\s+\{\s+samples:\s+(\d+),/ && ! ($opt_m || $opt_s) ) {
	    my $type = $1;
	    my $count = $2;
	    if ( $opt_t && $opt_t ne $type ) {
		# do nothing since type does not match
		print "Found line with samples, type ($type) and count ($count) but type ($opt_t) does not match: $line\n" if($opt_d);
	    } else {
		if ( $job_type2count{$jobid}{$type} ) {
		    $job_type2count{$jobid}{$type} += $count;
		} else {
		    $job_type2count{$jobid}{$type} = $count;
		}
	    }
	    print "Found line with samples, type ($type) and count ($count): $line\n" if($opt_d);
	} elsif ( $line =~ /^\s+(\S+):\s+\{\s+samples:\s+(\d+),\s+unit:\s+bytes,.*,\s+sum:\s+(\d+)/ && ($opt_m || $opt_s) ) {
	    my $type = $1;
	    my $count = $2;
	    my $sum = $3;
	    if ( $opt_t && $opt_t ne $type ) {
		# do nothing since type does not match
		print "Found line with sum, type ($type), count ($count) and sum ($sum) but type ($opt_t) does not match: $line\n" if($opt_d);
	    } else {
		if ( $job_type2count{$jobid}{$type} ) {
		    $job_type2count{$jobid}{$type} += $count;
		} else {
		    $job_type2count{$jobid}{$type} = $count;
		}
		if ( $job_type2sum{$jobid}{$type} ) {
		    $job_type2sum{$jobid}{$type} += $sum;
		} else {
		    $job_type2sum{$jobid}{$type} = $sum;
		}
		print "Found line with sum, type ($type), count ($count) and sum ($sum): $line\n" if($opt_d);
	    }
	} elsif ( $line =~ /^\s+(\S+):\s+\{\s+samples:\s+(\d+),/ && ($opt_m || $opt_s) ) {
	    print "Found line without importance: $line\n" if($opt_d);
	} elsif ( $line =~ /^\s+snapshot_time:\s+\d+/ || $line =~ /job_stats:\s*$/ || $line =~ /job_stats=\s*$/ ) {
	    print "Found line without importance: $line\n" if($opt_d);
	} else {
	    print STDERR "Exiting because of unexpected line: $line\n";
	    exit 1;
	}
    }
}


# from Perl FAQ: add commas to numbers (makes huge numbers better readable)
sub commify {
    my $input = shift;
    $input = reverse $input;
    $input =~ s/(\d\d\d)(?=\d)(?!\d*\.)/$1,/g;
    return reverse $input;
}


sub create_output {
    my $medsize = 0;
    my @joblist;
    foreach my $jobid (sort(keys(%job_type2count))) {
	foreach my $type (keys %{ $job_type2count{$jobid} }) {
	    if ( $opt_m ) {
		if ( $job_type2count{$jobid}{$type} > 0 ) {
		    $medsize = int( $job_type2sum{$jobid}{$type} / $job_type2count{$jobid}{$type} );
		} elsif ( $job_type2sum{$jobid}{$type} > 0 ) {
		    print STDERR "Job $jobid has sum > 0 ($job_type2sum{$jobid}{$type}) but no count > 0 ($job_type2count{$jobid}{$type}). Exiting...\n";
		    exit 1;
		}
	    }
	    # Just check if any requirement is *not* fulfilled and in this case do nothing.
	    unless ( ($opt_m && $medsize > $opt_m) || ($opt_o && $job_type2count{$jobid}{$type} < $opt_o) || ($opt_s && $job_type2sum{$jobid}{$type} < $opt_s) ) {
		print "Job $jobid has done ";
		if ( $opt_o ) {
		    print "$job_type2count{$jobid}{$type} ";
		}
		print "$type operations";
		if ( $opt_m || $opt_s ) {
		    print " with ";
		}
		if ( $opt_m ) {
		    print "medium size $medsize bytes";
		}
		if ( $opt_m && $opt_s ) {
		    print " and ";
		}
		if ( $opt_s ) {
		    my $commanum = commify ( $job_type2sum{$jobid}{$type} );
		    print "sum of $commanum bytes";
		}
		print ".\n";
		push(@joblist, "$jobid");
	    }
	}
    }
    if ( join("", @joblist) ) {
	print "List of found job IDs: " . join(" ",uniq sort(@joblist)) . "\n";
    }
}


# get options
getopts("dhvm:o:s:t:") || usage;
usage if ( $opt_h );
usage unless ($#ARGV == -1);

if ( ($opt_m && $opt_m !~ /^\d+$/) || ($opt_o && $opt_o !~ /^\d+$/) || ($opt_s && $opt_s !~ /^\d+$/) ) {
    print STDERR "Specify a number for options -m, -o or -s\n";
    exit 1;
}
unless ( $opt_m || $opt_o || $opt_s ) {
    print STDERR "One of the options -m, -o or -s is required\n";
    exit 1;
}

parse_input;

create_output;

exit 0;
